//#####################################################################
// Copyright 2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Math_Tools/RANGE.h>
#include "QT_CAMERA.h"
#include <QMatrix4x4>

namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
QT_CAMERA::
QT_CAMERA()
{
    rotation_matrix=TM::Identity_Matrix();
}
//#####################################################################
// Destructor
//#####################################################################
QT_CAMERA::
~QT_CAMERA()
{
}
//#####################################################################
// Function Set_Size
//#####################################################################
void QT_CAMERA::
Set_Size(int w, int h)
{
    width=w;
    height=h;
}
//#####################################################################
// Function Begin_Rotation
//#####################################################################
void QT_CAMERA::
Begin_Rotation(int x, int y)
{
    if(mode_2d) return;
    TV2 v(2.f*x/(width-1)-1,1-2.f*y/(height-1));
    arcball.Begin_Drag(v);
    do_rotation = true;
}
//#####################################################################
// Function Apply_Rotation
//#####################################################################
void QT_CAMERA::
Apply_Rotation(int x, int y)
{
    if(mode_2d) return;
    TV2 v(2.f*x/(width-1)-1,1-2.f*y/(height-1));
    arcball.Update(v);
}
//#####################################################################
// Function End_Rotation
//#####################################################################
void QT_CAMERA::
End_Rotation(int x, int y)
{
    if(mode_2d) return;
    rotation_matrix=arcball.rotation_matrix*rotation_matrix;
    arcball.End_Drag();
    do_rotation = false;
}
//#####################################################################
// Function Begin_Zoom
//#####################################################################
void QT_CAMERA::
Begin_Zoom(int x, int y)
{
    mouse_last_x=x;
    mouse_last_y=y;
    do_zoom = true;
}
//#####################################################################
// Function Apply_Zoom
//#####################################################################
void QT_CAMERA::
Apply_Zoom(int x, int y)
{
    float dy = 2.f*(mouse_last_y - y)/height;
    mouse_last_y=y;
    scale*=pow(8,dy);
}
//#####################################################################
// Function End_Zoom
//#####################################################################
void QT_CAMERA::
End_Zoom(int x, int y)
{
    do_zoom = false;
}
//#####################################################################
// Function Begin_Translation
//#####################################################################
void QT_CAMERA::
Begin_Translation(int x, int y)
{
    mouse_last_x=x;
    mouse_last_y=y;
    do_translation = true;
    auto m=Transform();
    auto stx=m*target_position.Append(1);
    stx/=stx(3);
    auto sty=stx;
    stx(0)+=2.f/width;
    sty(1)+=2.f/height;
    auto inx=m.Inverse_Times(stx);
    auto iny=m.Inverse_Times(sty);
    inx/=inx(3);
    iny/=iny(3);
    target_dx=inx.Remove_Index(3)-target_position;
    target_dy=iny.Remove_Index(3)-target_position;
}
//#####################################################################
// Function Apply_Translation
//#####################################################################
void QT_CAMERA::
Apply_Translation(int x, int y)
{
    target_position+=(mouse_last_x-x)*target_dx+(y-mouse_last_y)*target_dy;
    mouse_last_x=x;
    mouse_last_y=y;
}
//#####################################################################
// Function End_Translation
//#####################################################################
void QT_CAMERA::
End_Translation(int x, int y)
{
    do_translation = false;
}
//#####################################################################
// Function Transform
//#####################################################################
auto QT_CAMERA::
Transform() const -> TM4
{
    TM4 m=Perspective();
    TM4 t0,t1;
    t0.Set_Column(TV::m,TV4{0,0,-scale,0});
    m=m*(t0+1);
    if(!mode_2d)
    {
        TM4 r;
        r.Set_Submatrix(0,0,arcball.rotation_matrix*rotation_matrix);
        r(3,3)=1;
        m=m*r;
    }
    t1.Set_Column(TV::m,-target_position.Append(0));
    m=m*(t1+1);
    return m;
}
//#####################################################################
// Function Camera_Frame_To_World
//#####################################################################
auto QT_CAMERA::
Camera_Frame_To_World(const TV& u) const -> TV
{
    TV v=u.Add_Axis(2,scale);
    if(!mode_2d)
    {
        v=arcball.rotation_matrix.Transpose_Times(v);
        v=rotation_matrix.Transpose_Times(v);
    }
    return v+target_position;
}
//#####################################################################
// Function Look_At
//#####################################################################
void QT_CAMERA::
Look_At(const TV& X)
{
    target_position=X;
}
//#####################################################################
// Function Look_At
//#####################################################################
void QT_CAMERA::
Look_At(const RANGE<TV>& box)
{
    target_position=box.Center();
    auto size=box.Edge_Lengths();

    float y=max(size.x*height/width,size.y);
    scale=.5f*y/tan(.5f*fovy*(float)pi/180)+.5f*size.z;
    scale=max(scale,1e-4f);
}
//#####################################################################
// Function Perspective
//#####################################################################
auto QT_CAMERA::
Perspective() const -> TM4
{
    float aspect=(float)width/height;
    TM4 m;
    float nearclip=nearclip_factor*scale;
    float farclip=farclip_factor*scale;
    float f=1/tan(.5f*(float)pi/180*fovy);
    m(0,0)=f/aspect;
    m(1,1)=f;
    m(2,2)=(farclip+nearclip)/(nearclip-farclip);
    m(2,3)=(2*farclip*nearclip)/(nearclip-farclip);
    m(3,2)=-1;
    return m;
}
}
