//#####################################################################
// Copyright 2004-2009, Eran Guendelman, Frank Losasso, Andrew Selle.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Arrays/ARRAYS_ND.h>
#include <Core/Basic_Geometry/ORIENTED_BOX.h>
#include <Core/Log/LOG.h>
#include <Core/Matrices/FRAME.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include "../QT_MAIN_WINDOW.h"
#include "../QT_RENDER_UTILS.h"
#include "../QT_WORLD.h"
#include "QT_COMPONENT_OPTIONS.h"
#include "QT_GROUP.h"
#include "QT_PARTICLES.h"
#include <QMenuBar>
#include <sstream>
#include <stdexcept>
namespace PhysBAM{
#define SET(name,value) (opt.name.Set(value),raw_opt[0].name=opt.name)
#define GET(name) opt.name.Value()
//#####################################################################
// Constructor
//#####################################################################
QT_GROUP::
QT_GROUP(QT_WORLD& world,const std::string& name)
    :QT_COMPONENT(world,name),opt(OPT::def),raw_opt{{},{},{},OPT::def}
{
}
//#####################################################################
// Destructor
//#####################################################################
QT_GROUP::
~QT_GROUP()
{
}
//#####################################################################
// Function Print_Particle
//#####################################################################
template<class TV> bool
Print_Particle(std::ostream& o,QT_COMPONENT* obj,int id)
{
    if(auto p=dynamic_cast<QT_PARTICLES<TV>*>(obj))
    {
        p->Print_Particle(o,id);
        return true;
    }
    return false;
}
//#####################################################################
// Function Render
//#####################################################################
void QT_GROUP::
Render(QT_RENDER_UTILS& ru)
{
    if(!GET(draw)) return;
    Read();
    MATRIX<float,4> fr;
    fr.Set_Submatrix(0,0,frame.r.Rotation_Matrix());
    fr.Set_Column(3,frame.t.Append(1));
    auto old=ru.Push(ru.obj_to_world,fr*ru.obj_to_world);

    if(!ru.select && selected_component.size())
    {
        LOG::fprintf(ru.info_text,"\n");
        if(!GET(source).empty() && !GET(attr).empty() && GET(particle_id)>=0)
        {
            if(auto obj=world.Find_Component(GET(source)))
            {
                if(Print_Particle<VECTOR<float,2>>(ru.info_text,&*obj,GET(particle_id))){}
                else if(Print_Particle<VECTOR<double,2>>(ru.info_text,&*obj,GET(particle_id))){}
                else if(Print_Particle<VECTOR<float,3>>(ru.info_text,&*obj,GET(particle_id))){}
                else if(Print_Particle<VECTOR<double,3>>(ru.info_text,&*obj,GET(particle_id))){}
            }
        }
        LOG::fprintf(ru.info_text,"\ncomponent %P\n\n",selected_component);
    }
    
    ru.pick.Push(0);
    int k=0;
    for(const auto& a:components)
    {
        ru.pick.Set(k++);
        a.second->Render(ru);
    }
    ru.pick.Pop();
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
auto QT_GROUP::
Bounding_Box() -> RANGE<FV>
{
    if(!GET(draw)) return RANGE<FV>::Empty_Box();
    Read();
    RANGE<FV> box=RANGE<FV>::Empty_Box();
    for(const auto& a:components) box=box.Unite(a.second->Bounding_Box());
    if(box.Empty()) return box;
    return ORIENTED_BOX<FV>(box,frame).Axis_Aligned_Bounding_Box();
}
//#####################################################################
// Function Get_Selection_Priority
//#####################################################################
// int QT_GROUP::
// Get_Selection_Priority(ARRAY_VIEW<GLuint> indices)
// {
//     if(!indices.m) return -1;
//     return 100;
// }
//#####################################################################
// Function Get_Selection
//#####################################################################
void QT_GROUP::
Set_Selection(ARRAY_VIEW<int> selection)
{
    int k=selection(0);
    for(const auto&a:components)
        if(!k--)
        {
            selected_component=a.first;
            a.second->Set_Selection(selection.Array_View(1,selection.m-1));
        }
}
//#####################################################################
// Function Clear_Highlight
//#####################################################################
void QT_GROUP::
Clear_Selection()
{
    for(const auto&a:components) a.second->Clear_Selection();
}
//#####################################################################
// Function Get_Frame
//#####################################################################
template<class TV> bool
Get_Frame(QT_COMPONENT* obj,const std::string& attr,int id,FRAME<VECTOR<float,3>>& f)
{
    if(auto p=dynamic_cast<QT_PARTICLES<TV>*>(obj))
    {
        auto [ff,b]=p->Get_Frame(attr,id);
        if(b) f=ff;
        return b;
    }
    return false;
}
//#####################################################################
// Function Reinitialize
//#####################################################################
void QT_GROUP::
Read()
{
    for(const auto& a:components)
        a.second->Read();

    frame=GET(frame);
    if(!GET(source).empty() && !GET(attr).empty() && GET(particle_id)>=0)
    {
        if(auto obj=world.Find_Component(GET(source)))
        {
            obj->Read();
            if(Get_Frame<VECTOR<float,2>>(&*obj,GET(attr),GET(particle_id),frame)){}
            else if(Get_Frame<VECTOR<double,2>>(&*obj,GET(attr),GET(particle_id),frame)){}
            else if(Get_Frame<VECTOR<float,3>>(&*obj,GET(attr),GET(particle_id),frame)){}
            else if(Get_Frame<VECTOR<double,3>>(&*obj,GET(attr),GET(particle_id),frame)){}
        }
    }
}
//#####################################################################
// Function Invalidate
//#####################################################################
void QT_GROUP::
Invalidate(bool invalidate_common)
{
    for(const auto& a:components)
        a.second->Invalidate(invalidate_common);
}
//#####################################################################
// Function Selection_Point
//#####################################################################
auto QT_GROUP::
Selection_Point() -> FV
{
    return frame*components[selected_component]->Selection_Point();
}
//#####################################################################
// Function Add_To_Menu
//#####################################################################
void QT_GROUP::
Add_To_Menu(QT_MAIN_WINDOW& main,QMenu* menu)
{
    auto submenu = menu->addMenu(name.c_str());
    auto action = submenu->addAction("[group]");
    action->setCheckable(true);
    action->setChecked(Draw());
    QObject::connect(action,&QAction::triggered,&main,
        [this,&main](){main.Options_Dialog(this);});
    for(auto& c:components)
        c.second->Add_To_Menu(main,submenu);
}
//#####################################################################
// Function Make_Dialog
//#####################################################################
void QT_GROUP::
Make_Dialog(QT_WINDOW& window)
{
    QT_COMPONENT_OPTIONS* dialog=new QT_COMPONENT_OPTIONS(window,*this);
    dialog->setWindowTitle(QString::fromStdString(name));
    OPT opt2,raw_opt2;
    opt2=opt;
    raw_opt2=raw_opt[0];
    dialog->Add("draw",opt.draw,raw_opt[0].draw);
    dialog->Add("source",opt.source,raw_opt[0].source);
    dialog->Add("attr",opt.attr,raw_opt[0].attr);
    dialog->Add("particle_id",opt.particle_id,raw_opt[0].particle_id);
    dialog->Add_End_Buttons();
    int result=dialog->exec();
    if(result!=QDialog::Accepted)
    {
        opt=opt2;
        raw_opt[0]=raw_opt2;
    }
    window.update();
}
//#####################################################################
// Function Read_Option
//#####################################################################
void QT_GROUP::
Read_Option(std::istream& in,int level)
{
    auto& o=raw_opt[level];
    o.Read_Entry(in);
    opt=raw_opt[3];
    opt.Merge(raw_opt[2]);
    opt.Merge(raw_opt[1]);
    opt.Merge(raw_opt[0]);
    auto p=world.Find_Component(this->name);
    PHYSBAM_ASSERT(p);
    for(const auto&s:GET(add))
    {
        auto c=world.Find_Component(s);
        if(c)
        {
            if(c->parent) continue;
            c->parent=p;
            components[s]=c;
        }
    }
    GET(add).Remove_All();
}
//#####################################################################
// Function Set_Draw
//#####################################################################
void QT_GROUP::
Set_Draw(bool draw)
{
    SET(draw,draw);
    action->setChecked(Draw());
}
//#####################################################################
// Function Draw
//#####################################################################
bool QT_GROUP::
Draw() const
{
    return GET(draw);
}
//#####################################################################
// Function Save_Settings
//#####################################################################
void QT_GROUP::
Save_Settings(std::ostream& out)
{
    raw_opt[0].Write(out,name);
    for(const auto& a:components)
        a.second->Save_Settings(out);
}
//#####################################################################
// Function Clear_Settings
//#####################################################################
void QT_GROUP::
Clear_Settings()
{
    raw_opt[0]={};
    opt=raw_opt[3];
    opt.Merge(raw_opt[2]);
    opt.Merge(raw_opt[1]);
    opt.Merge(raw_opt[0]);
    for(const auto& a:components)
        a.second->Clear_Settings();
}
}
