#ifndef __GAUSS_QUADRATURE__
#define __GAUSS_QUADRATURE__

#include <Core/Math_Tools/RANGE.h>
#include <Core/Vectors/VECTOR.h>
#include <functional>
#include <iostream>
#include <utility>


namespace PhysBAM {

// From https://pomax.github.io/bezierinfo/legendre-gauss.html
//
// Gaussian Quadrature weights and locations
//
// Each pair is a (weight, location) pair
// The n-th element of the outer array gives the weights and locations for that order.
static constexpr std::pair<double, double> QUADRATURE_WEIGHTS[][7] = 
{
    {}, // n = 0
    {   // n = 1
        {2.0, 0}
    },
    {   // n = 2
        {1.0000000000000000, -0.5773502691896257}, 
        {1.0000000000000000, 0.5773502691896257}
    },
    {   // n = 3
        {0.8888888888888888, 0.0000000000000000},
        {0.5555555555555556, -0.7745966692414834},
        {0.5555555555555556, 0.7745966692414834}
    },
    {   // n = 4
        {0.6521451548625461,-0.3399810435848563},
        {0.6521451548625461, 0.3399810435848563},
        {0.3478548451374538,-0.8611363115940526},
        {0.3478548451374538, 0.8611363115940526}
    },
    {   // n = 5
        {0.5688888888888889, 0.0000000000000000},
        {0.4786286704993665,-0.5384693101056831},
        {0.4786286704993665, 0.5384693101056831},
        {0.2369268850561891,-0.9061798459386640},
        {0.2369268850561891, 0.9061798459386640}
    },
    {   // n = 6
        {0.3607615730481386, 0.6612093864662645},
        {0.3607615730481386,-0.6612093864662645},
        {0.4679139345726910,-0.2386191860831969},
        {0.4679139345726910, 0.2386191860831969},
        {0.1713244923791704,-0.9324695142031521},
        {0.1713244923791704, 0.9324695142031521}
    },
};

// Uses Gaussian-Quadrature to integrate "f" on [-1, 1]
double Gauss_Quadrature(std::function<double(double)> f, int order)
{
    auto& weights=QUADRATURE_WEIGHTS[order];
    double sum=0;
    for (int i=0;i<order;i++)
        sum += weights[i].first * f(weights[i].second);
    return sum;
}

double Gauss_Quadrature(std::function<double(double)> f, double a, double b, int order)
{

    return 0.5*(b-a)*Gauss_Quadrature(
        [&f, a, b](double x)
        {
            double u=0.5*(b-a)*(x+1) + a;
            return f(u);
        },
        order);
}

// Uses Gaussian-Quadrature to integrate "f" on [-1, 1] x [-1, 1]
template<class T>
T Gauss_Quadrature_2D(std::function<T(VECTOR<T, 2>)> f, int order)
{
    typedef VECTOR<T,2> TV;

    auto& weights=QUADRATURE_WEIGHTS[order];
    double sum=0;
    for(int i=0;i<order;i++){
        for(int j=0;j<order;j++){
            sum += weights[i].first * weights[j].first * f(TV(weights[i].second, weights[j].second));
        }
    }
    return sum;
}

template<class T>
T Gauss_Quadrature_2D(std::function<T(VECTOR<T, 2>)> f, RANGE<VECTOR<T, 2>> domain, int order) {
    typedef VECTOR<T,2> TV;

    TV edges=domain.Edge_Lengths();
    return 0.25*edges.Product()*Gauss_Quadrature_2D<T>(
        [&f, &edges, &domain](TV X)
        {
            TV u=0.5*edges*(X+TV::All_Ones_Vector()) + domain.min_corner;
            return f(u);
        },
        order);
}

}

#endif
