//#####################################################################
// Copyright 2019, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Utilities/VIEWER_DIR.h>
#include <Core/Arrays/LEXICOGRAPHIC_COMPARE.h>
#include <Core/Log/LOG.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <fstream>
namespace PhysBAM{
//#####################################################################
// Constructor
//#####################################################################
VIEWER_DIR::
VIEWER_DIR(const std::string& output_directory)
    :output_directory(output_directory)
{
}
//#####################################################################
// Destructor
//#####################################################################
VIEWER_DIR::
~VIEWER_DIR()
{
}
//#####################################################################
// Function Advance_Directory
//#####################################################################
void VIEWER_DIR::
Advance_Directory(int substep_level)
{
    if(!frame_stack.m)
    {
        assert(substep_level==0);
        frame_stack.Append(0);
    }
    else
    {
        assert((unsigned)substep_level<=max_substep_level);
        frame_stack.Resize(substep_level+1);
        frame_stack.Last()++;
    }
    Update_Current_Directory();
}
//#####################################################################
// Function Update_Current_Directory
//#####################################################################
void VIEWER_DIR::
Update_Current_Directory()
{
    current_directory=Current_Directory(frame_stack);
}
//#####################################################################
// Function Update_Current_Directory
//#####################################################################
std::string VIEWER_DIR::
Current_Directory(const ARRAY<int>& stack) const
{
    if(!frame_stack.m) return "";
    return output_directory+"/"+Frame_As_String(stack);
}
//#####################################################################
// Function Frame_As_String
//#####################################################################
std::string VIEWER_DIR::
Frame_As_String() const
{
    return Frame_As_String(frame_stack);
}
//#####################################################################
// Function Frame_As_String
//#####################################################################
std::string VIEWER_DIR::
Frame_As_String(const ARRAY<int>& stack) const
{
    if(!stack.m) return "";
    std::string s=std::to_string(stack(0));
    for(int i=1;i<stack.m;i++)
    {
        s+=".";
        s+=std::to_string(stack(i));
    }
    return s;
}
//#####################################################################
// Function Make_Common_Directory
//#####################################################################
void VIEWER_DIR::
Make_Common_Directory(bool append_log)
{
    if(made_common) return;
    fs::create_directories(output_directory);
    fs::create_directories(output_directory+"/common");
    if(!no_log)
        LOG::Instance()->Copy_Log_To_File(Common_File("log.txt"),append_log);
    made_common=true;
}
//#####################################################################
// Function Start_Directory
//#####################################################################
void VIEWER_DIR::
Start_Directory(int substep_level,const char* title)
{
    Advance_Directory(substep_level);
    if(First_Frame()) Make_Common_Directory();
    fs::create_directories(current_directory);
    if(title) Write_To_Text_File(current_directory+"/frame_title",title);
}
//#####################################################################
// Function Start_Directory
//#####################################################################
void VIEWER_DIR::
Start_Directory(int substep_level,const std::string& title)
{
    Start_Directory(substep_level,title.c_str());
}
//#####################################################################
// Function Finish_Directory
//#####################################################################
void VIEWER_DIR::
Finish_Directory()
{
    std::ofstream lf(Common_File("last_frame"));
    frame_stack.Write_Raw(lf);
    lf<<std::endl;
}
//#####################################################################
// Function Find_Next_Directory
//#####################################################################
bool VIEWER_DIR::
Find_Next_Directory(int substep_level,bool check_last_frame)
{
    SAVE_STATE save;
    Save_State(save);
    ARRAY<int> last_frame_stack;
    if(check_last_frame) Read_Last_Frame(last_frame_stack);
    assert(substep_level>=0 && substep_level<=max_substep_level);
    for(int i=substep_level;i>=0;i--)
    {
        Advance_Directory(i);
        if(check_last_frame && LEXICOGRAPHIC_COMPARE()(last_frame_stack,frame_stack))
            break;
        if(fs::is_directory(current_directory))
            return true;
    }
    Restore_State(save);
    return false;
}
//#####################################################################
// Function Valid_Frame
//#####################################################################
bool VIEWER_DIR::
Valid_Frame(const ARRAY<int>& stack, bool check_last_frame) const
{
    if(check_last_frame){
        ARRAY<int> last_frame_stack;
        Read_Last_Frame(last_frame_stack);
        if(LEXICOGRAPHIC_COMPARE()(last_frame_stack,stack))
            return false;}
    return fs::is_directory(Current_Directory(stack));
}
//#####################################################################
// Function Valid_Frame
//#####################################################################
bool VIEWER_DIR::
Valid_Frame(bool check_last_frame) const
{
    if(check_last_frame){
        ARRAY<int> last_frame_stack;
        Read_Last_Frame(last_frame_stack);
        if(LEXICOGRAPHIC_COMPARE()(last_frame_stack,frame_stack))
            return false;}
    return fs::is_directory(current_directory);
}
//#####################################################################
// Function Find_Prev_Directory
//#####################################################################
bool VIEWER_DIR::
Find_Prev_Directory(int substep_level,bool check_last_frame)
{
    PHYSBAM_ASSERT(substep_level>=0);
    if(!frame_stack.m) return false;
    if(frame_stack.m==1 && frame_stack(0)==0) return false;
    if(check_last_frame)
    {
        ARRAY<int> last_frame_stack;
        Read_Last_Frame(last_frame_stack);
        if(LEXICOGRAPHIC_COMPARE()(last_frame_stack,frame_stack))
        {
            frame_stack=last_frame_stack;
            if(frame_stack.m<=substep_level+1 && Valid_Frame(false))
            {
                Update_Current_Directory();
                return true;
            }
        }
    }
    if(frame_stack.m>substep_level+1)
    {
        frame_stack.Resize(substep_level+1);
        while(frame_stack.m>1 && frame_stack.Last()==0) frame_stack.Pop();
        Update_Current_Directory();
        return true;
    }
    while(frame_stack.m>1 && frame_stack.Last()==0) frame_stack.Pop();
    frame_stack.Last()--;
    Update_Current_Directory();
    for(int s=frame_stack.m;s<=substep_level;s++)
    {
        frame_stack.Append(0);
        while(fs::is_directory(current_directory+"."+std::to_string(frame_stack.Last()+1)))
            frame_stack.Last()++;
        current_directory+="."+std::to_string(frame_stack.Last());
    }
    while(frame_stack.m>1 && !frame_stack.Last()) frame_stack.Pop();
    Update_Current_Directory();
    return true;
}
//#####################################################################
// Function Set
//#####################################################################
void VIEWER_DIR::
Set(int frame)
{
    frame_stack.Resize(1);
    frame_stack(0)=frame;
    current_directory=output_directory+"/"+std::to_string(frame_stack(0));
}
//#####################################################################
// Function Set
//#####################################################################
void VIEWER_DIR::
Set(const char* frame_string)
{
    Parse_Frame(frame_stack, frame_string);
    Update_Current_Directory();
}
//#####################################################################
// Function Set
//#####################################################################
void VIEWER_DIR::
Set(const std::string& frame_string)
{
    Set(frame_string.c_str());
}
//#####################################################################
// Function Read_Last_Frame
//#####################################################################
void VIEWER_DIR::
Read_Last_Frame(ARRAY<int>& stack) const
{
    std::string line;
    std::ifstream lf(Common_File("last_frame"));
    getline(lf,line);
    Parse_Frame(stack, line);
}
//#####################################################################
// Function Read_Last_Frame
//#####################################################################
void VIEWER_DIR::
Read_Last_Frame(int substep_level)
{
    Read_Last_Frame(frame_stack);
    if(frame_stack.m>substep_level+1)
        frame_stack.Resize(substep_level+1);
    Update_Current_Directory();
}
//#####################################################################
// Function Parse_Frame
//#####################################################################
void VIEWER_DIR::
Parse_Frame(ARRAY<int>& stack,const std::string& frame_string) const
{
    Parse_Frame(stack,frame_string.c_str());
}
//#####################################################################
// Function Parse_Frame
//#####################################################################
void VIEWER_DIR::
Parse_Frame(ARRAY<int>& stack, const char* frame_string) const
{
    stack.Remove_All();
    char* t=0;
    while(1)
    {
        frame_string+=strcspn(frame_string,"0123456789");
        if(!*frame_string) break;
        stack.Append(strtol(frame_string, &t, 10));
        frame_string=t;
    }
}
//#####################################################################
// Function Save_State
//#####################################################################
void VIEWER_DIR::
Save_State(SAVE_STATE& s) const
{
    s.current_directory=current_directory;
    s.frame_stack=frame_stack;
    s.made_common=made_common;
}
//#####################################################################
// Function Restore_State
//#####################################################################
void VIEWER_DIR::
Restore_State(const SAVE_STATE& s)
{
    current_directory=s.current_directory;
    frame_stack=s.frame_stack;
    made_common=s.made_common;
}
//#####################################################################
// Function Common_File
//#####################################################################
std::string VIEWER_DIR::
Common_File(const char* filename) const
{
    return output_directory+"/common/"+filename;
}
//#####################################################################
// Function Common_File
//#####################################################################
std::string VIEWER_DIR::
Common_File(const std::string& filename) const
{
    return Common_File(filename.c_str());
}
//#####################################################################
// Function Frame_File
//#####################################################################
std::string VIEWER_DIR::
Frame_File(const char* filename) const
{
    return current_directory+"/"+filename;
}
//#####################################################################
// Function Frame_File
//#####################################################################
std::string VIEWER_DIR::
Frame_File(const std::string& filename) const
{
    return Frame_File(filename.c_str());
}
//#####################################################################
// Function First_Frame
//#####################################################################
bool VIEWER_DIR::
First_Frame() const
{
    return frame_stack.m==1 && frame_stack(0)==0;
}
}
